from typing import Any, List, Optional

from cfinterface.components.field import Field


class Line:
    """
    Class for representing a generic line that is composed
    of fields and can be read from or written to a file.
    """

    def __init__(
        self, fields: List[Field], values: Optional[List[Any]] = None
    ):
        self._fields = fields
        if values is not None:
            for f, v in zip(self._fields, values):
                f.value = v

    def read(self, line: str) -> List[Any]:
        """
        Reads a line for extracting information following
        the given fields.

        :param line: The line to be read
        :type line: str
        :return: The extracted values, in order
        :rtype: List[Any]
        """
        for field in self._fields:
            field.read(line)
        return self.values

    def write(self, line: str) -> str:
        """
        Writes to a line with the existing information
        if the given fields.

        :param line: The line to be written
        :type line: str
        :return: The line with the new field information
        :rtype: str
        """
        newline = line
        for field in self._fields:
            newline = field.write(newline)
        return newline

    @property
    def values(self) -> List[Any]:
        return [f.value for f in self._fields]

    @values.setter
    def values(self, vals: List[Any]):
        for f, v in zip(self._fields, vals):
            f.value = v



from typing import IO, List

from cfinterface.components.block import Block
from cfinterface.components.state import ComponentState
from cfinterface.data.blockdata import BlockData
from cfinterface.files.blockfile import BlockFile

from tests.mocks.mock_open import mock_open

from unittest.mock import MagicMock, patch


class DummyBlock(Block):
    BEGIN_PATTERN = "beg"
    END_PATTERN = "end"

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, self.__class__):
            return False
        else:
            return o.data == self.data

    def read(self, file: IO) -> bool:
        self.data: List[str] = []
        while True:
            line: str = file.readline()
            self.data.append(line)
            if self.ends(line):
                break
        return True

    def write(self, file: IO) -> bool:
        for line in self.data:
            file.write(line)
        return True


def test_blockfile_read():
    data = "Hello, world!"
    filedata = (
        "\n".join([DummyBlock.BEGIN_PATTERN, data, DummyBlock.END_PATTERN])
        + "\n"
    )
    BlockFile.BLOCKS = [DummyBlock]
    m: MagicMock = mock_open(read_data=filedata)
    with patch("builtins.open", m):
        f = BlockFile.read("", "")
        assert len(f.data) == 2
        assert len(f.data.last.data) == 3
        assert f.data.last.data[1] == data + "\n"


def test_blockfile_write():
    data = "Hello, world!"
    bd = BlockData(DummyBlock(state=ComponentState.READ_SUCCESS, data=[data]))
    BlockFile.BLOCKS = [DummyBlock]
    f = BlockFile(bd)
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        f.write("", "")
    m().write.assert_called_once_with(data)



from abc import abstractmethod
from typing import Any, IO
import re

from cfinterface.components.state import ComponentState


class Block:
    """
    Class for a generic block in a textfile, with given markers
    for beginning and end and reading states.
    """

    BEGIN_PATTERN = ""
    END_PATTERN = ""
    MAX_LINES = 10000

    def __init__(
        self,
        state=ComponentState.NOT_FOUND,
        previous=None,
        next=None,
        data=None,
    ) -> None:
        self.__state = state
        self.__previous = previous
        self.__next = next
        self.__data: Any = data

    def __eq__(self, o: object) -> bool:
        raise NotImplementedError()

    @classmethod
    def begins(cls, line: str):
        """
        Checks if the current line marks the beginning of the block.

        :param line: The candidate line for being the beggining of
            the block.
        :type line: str
        """
        return re.search(cls.BEGIN_PATTERN, line) is not None

    @classmethod
    def ends(cls, line: str):
        """
        Checks if the current line marks the end of the block.

        :param line: The candidate line for being the end of the block.
        :type line: str
        """
        return re.search(cls.END_PATTERN, line) is not None

    @abstractmethod
    def read(self, file: IO) -> bool:
        """
        Generic function to perform the reading of the block using
        a filepointer.

        :param file: The filepointer
        :type file: IO
        :return: The success, or not, in the reading
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def write(self, file: IO) -> bool:
        """
        Generic function to perform the writing of the block using
        a filepointer.

        :param file: The filepointer
        :type file: IO
        :return: The success, or not, in the writing
        :rtype: bool
        """
        raise NotImplementedError()

    def read_block(self, file: IO):
        """
        Function that reads the block and evaluates the result.

        :param file: The filepointer
        :type file: IO
        """
        if self.read(file):
            self.__state = ComponentState.READ_SUCCESS
        else:
            self.__state = ComponentState.READ_ERROR

    def write_block(self, file: IO):
        """
        Function that writes the block, if it was succesfully read.

        :param file: The filepointer
        :type file: IO
        """
        if self.__state == ComponentState.READ_SUCCESS:
            if self.write(file):
                self.__state = ComponentState.WRITE_SUCCESS
            else:
                self.__state = ComponentState.WRITE_ERROR

    @property
    def previous(self) -> "Block":
        return self.__previous

    @previous.setter
    def previous(self, b: "Block"):
        self.__previous = b

    @property
    def next(self) -> "Block":
        return self.__next

    @next.setter
    def next(self, b: "Block"):
        self.__next = b

    @property
    def data(self) -> Any:
        return self.__data

    @data.setter
    def data(self, d: Any):
        self.__data = d

    @property
    def is_first(self) -> bool:
        return self.__previous is None

    @property
    def is_last(self) -> bool:
        return self.__next is None

    @property
    def empty(self) -> bool:
        return self.__data is None

    @property
    def success(self) -> bool:
        return self.__state in [
            ComponentState.READ_SUCCESS,
            ComponentState.WRITE_SUCCESS,
        ]



from typing import Optional

from cfinterface.components.field import Field


class FloatField(Field):
    """
    Class for representing an float field for being read from and
    written to a file. The format to read and write the value is given
    by 'F' for fixed point notation and 'E' for scientific notation.
    """

    def __init__(
        self,
        size: int,
        starting_column: int,
        decimal_digits: int,
        format: str = "F",
        value: Optional[float] = None,
    ) -> None:
        super().__init__(size, starting_column, value)
        self.__decimal_digits = decimal_digits
        self.__format = format

    # Override
    def read(self, line: str) -> Optional[float]:
        linevalue = line[self._starting_column : self._ending_column].strip()
        self._value = (
            float(linevalue)
            if linevalue.replace(".", "").replace(self.__format, "").isdigit()
            else None
        )
        return self._value

    # Override
    def write(self, line: str) -> str:
        if self.value is None:
            raise ValueError("Field cannot be written if has no value")
        if len(line) < self._ending_column:
            line = line.ljust(self._ending_column)
        value = ""
        for d in range(self.__decimal_digits, -1, -1):
            value = "{:.{d}{format}}".format(
                round(self.value, d),
                d=d,
                format=self.__format,
            )
            if len(value) <= self._size:
                break

        return (
            line[: self._starting_column]
            + value.rjust(self._size)
            + line[self._ending_column :]
        )

    @property
    def value(self) -> Optional[float]:
        return self._value

    @value.setter
    def value(self, val: float):
        self._value = val



import pytest

from cfinterface.components.field import Field


def test_default_field():
    field = Field(10, 20, value=1)
    assert field.value == 1


def test_default_field_read_error():
    with pytest.raises(NotImplementedError):
        field = Field(10, 20)
        field.read("")


def test_default_field_write_error():
    with pytest.raises(NotImplementedError):
        field = Field(10, 20, value=1)
        field.write("")



from typing import Optional

from cfinterface.components.field import Field


class IntegerField(Field):
    """
    Class for representing an integer field for being read from and
    written to a file.
    """

    def __init__(
        self, size: int, starting_column: int, value: Optional[int] = None
    ) -> None:
        super().__init__(size, starting_column, value)

    # Override
    def read(self, line: str) -> Optional[int]:
        linevalue = line[self._starting_column : self._ending_column].strip()
        self._value = int(linevalue) if linevalue.isdigit() else None
        return self._value

    # Override
    def write(self, line: str) -> str:
        if self.value is None:
            raise ValueError("Field cannot be written if has no value")
        if len(line) < self._ending_column:
            line = line.ljust(self._ending_column)
        return (
            line[: self._starting_column]
            + str(self.value).rjust(self._size)
            + line[self._ending_column :]
        )

    @property
    def value(self) -> Optional[int]:
        return self._value

    @value.setter
    def value(self, val: int):
        self._value = val



from cfinterface.components.integerfield import IntegerField

import pytest


def test_integerfield_read():
    data = 12345
    field = IntegerField(5, 0)
    line = f"{data}-something-else"
    field.read(line)
    assert field.value == data


def test_integerfield_write():
    line_before = "field-12345-else"
    data = int(line_before[6:11])
    field = IntegerField(5, 6, value=data)
    line_after = field.write(line_before)
    assert line_before == line_after


def test_integerfield_write_error():
    with pytest.raises(ValueError):
        field = IntegerField(5, 0)
        field.write("")


def test_integerfield_write_short_line():
    line_before = "field-12345-else"
    data = int(line_before[6:11])
    field = IntegerField(5, 6, value=data)
    line_after = field.write("   ")
    assert str(data) == line_after[6:]



from typing import IO, List

from cfinterface.components.block import Block
from cfinterface.data.blockdata import BlockData
from cfinterface.writing.blockwriting import BlockWriting

from tests.mocks.mock_open import mock_open

from unittest.mock import MagicMock, patch


class DummyBlock(Block):
    BEGIN_PATTERN = "beg"
    END_PATTERN = "end"

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, self.__class__):
            return False
        else:
            return o.data == self.data

    def read(self, file: IO) -> bool:
        self.data: List[str] = []
        while True:
            line: str = file.readline()
            self.data.append(line)
            if self.ends(line):
                break
        return True

    def write(self, file: IO) -> bool:
        file.write(self.data)
        return True


def test_blockwriting_withdata():
    filedata = "Hello, World!"
    bd = BlockData(DummyBlock(data=filedata))
    bw = BlockWriting(bd)
    m: MagicMock = mock_open(read_data=filedata)
    with patch("builtins.open", m):
        bw.write("", "")
    m().write.assert_called_once_with(filedata)



from cfinterface.components.defaultblock import DefaultBlock
from cfinterface.components.state import ComponentState
from tests.mocks.mock_open import mock_open

from unittest.mock import MagicMock, patch


def test_default_block_eq():
    b1 = DefaultBlock(data=0)
    b2 = DefaultBlock(data=0)
    assert b1 == b2
    b1.data = 1
    assert b1 != b2
    b1.data = 0
    b1.previous = b2
    assert b1 != b2
    b1.previous = None
    b1.next = b2
    assert b1 != b2


def test_default_block_read():
    data = "Hello, world!\n"
    m: MagicMock = mock_open(read_data=data)
    with patch("builtins.open", m):
        with open("", "r") as fp:
            b = DefaultBlock()
            b.read_block(fp)
            assert b.data == data
            assert b.success


def test_default_block_write():
    data = "Hello, world!"
    filedata = ""
    m = mock_open(read_data=filedata)
    with patch("builtins.open", m):
        with open("", "w") as fp:
            b = DefaultBlock(state=ComponentState.READ_SUCCESS)
            b.data = data
            b.write_block(fp)
    m().write.assert_called_once_with(data)



from typing import TypeVar, Type, Generator

from cfinterface.components.block import Block


class BlockData:
    """
    Class for a storing, managing and accessing data for a block file.
    """

    T = TypeVar("T")

    def __init__(self, root: Block) -> None:
        self.__root = root
        self.__head = root

    def __iter__(self):
        current = self.__root
        while current:
            yield current
            current = current.next

    def __len__(self) -> int:
        count = 0
        for _ in self:
            count += 1
        return count

    def preppend(self, b: Block):
        """
        Appends a block to the beginning of the data.

        :param b: The new block to preppended to the data.
        :type b: Block
        """
        self.add_before(self.__root, b)

    def append(self, b: Block):
        """
        Appends a block to the end of the data.

        :param b: The new block to append to the data
        :type b: Block
        """
        self.add_after(self.__head, b)

    def add_before(self, before: Block, new: Block):
        """
        Adds a new block to the data before another
        specified block.

        :param before: The existing block which will be preppended
        :type before: Block
        :param new: The new block to add to the data
        :type new: Block
        """
        if before == self.__root:
            self.__root = new
        new.previous = before.previous
        before.previous = new
        new.next = before

    def add_after(self, after: Block, new: Block):
        """
        Adds a new block to the data after another
        specified block.

        :param after: The existing block which will be appended
        :type after: Block
        :param new: The new block to add to the data
        :type new: Block
        """
        if after == self.__head:
            self.__head = new
        new.next = after.next
        after.next = new
        new.previous = after

    def remove(self, b: Block):
        """
        Removes an existing block in the chain.

        :param b: The block to be removed
        :type b: Block
        """
        if b.previous is not None:
            b.previous.next = b.next
        if b.next is not None:
            b.next.previous = b.previous

    def of_type(self, t: Type[T]) -> Generator[T, None, None]:
        """
        A block generator that only returns blocks of type T.

        :param t: The block type that is desired
        :type t: Type[T]
        :yield: Blocks filtered by type T
        :rtype: Generator[T, None, None]
        """
        for b in self:
            if isinstance(b, t):
                yield b

    @property
    def first(self) -> Block:
        return self.__root

    @property
    def last(self) -> Block:
        return self.__head



import pytest

from cfinterface.components.literalfield import LiteralField


def test_literalfield_read():
    data = "field"
    field = LiteralField(len(data), 0)
    line = f"{data}-something-else"
    field.read(line)
    assert field.value == data


def test_literalfield_write():
    line_before = "field-something-else"
    data = line_before[6:15]
    field = LiteralField(len(data), 6, value=data)
    line_after = field.write(line_before)
    assert line_before == line_after


def test_literalfield_write_error():
    with pytest.raises(ValueError):
        field = LiteralField(5, 0)
        field.write("")


def test_literalfield_write_short_line():
    line_before = "field-something-else"
    data = line_before[6:15]
    field = LiteralField(len(data), 6, value=data)
    line_after = field.write("   ")
    assert data == line_after[6:]



from typing import IO
from os.path import join

from cfinterface.data.blockdata import BlockData


class BlockWriting:
    """
    Class for writing custom files based on a BlockData structure.
    """

    def __init__(self, data: BlockData) -> None:
        self.__data = data

    def __write_file(self, file: IO):
        """
        Writes all the blocks from the given BlockData structure
        to the specified file.

        :param file: The filepointer
        :type file: IO
        """
        for b in self.__data:
            b.write(file)

    def write(self, filename: str, directory: str):
        """
        Writes a file with a given name in a given directory with
        the data from the BlockData structure.

        :param filename: The name of the file
        :type filename: str
        :param directory: The directory where the file will be
        :type directory: str
        """
        filepath = join(directory, filename)
        with open(filepath, "w") as fp:
            return self.__write_file(fp)

    @property
    def data(self) -> BlockData:
        return self.__data



"""
cfi
====

cfi is a Python module for handling custom formatted files
and provide reading, storing and writing utilities.
"""

__version__ = "0.0.4"

from cfinterface.components.literalfield import LiteralField  # noqa
from cfinterface.components.integerfield import IntegerField  # noqa
from cfinterface.components.floatfield import FloatField  # noqa
from cfinterface.components.line import Line  # noqa
from cfinterface.components.defaultblock import DefaultBlock  # noqa
from cfinterface.data.blockdata import BlockData  # noqa
from cfinterface.reading.blockreading import BlockReading  # noqa
from cfinterface.writing.blockwriting import BlockWriting  # noqa
from cfinterface.files.blockfile import BlockFile  # noqa



from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField


def test_line_read_no_fields():
    line = Line([])
    fileline = ""
    assert len(line.read(fileline)) == 0


def test_line_read_with_fields():
    fields = [LiteralField(6, 0), LiteralField(6, 7)]
    line = Line(fields)
    fileline = "hello, world!"
    values = line.read(fileline)
    assert values[0] == "hello,"
    assert values[1] == "world!"


def test_line_write_no_fields():
    line = Line([])
    fileline = ""
    assert len(line.write(fileline)) == 0


def test_line_write_with_fields():
    fields = [LiteralField(6, 0), LiteralField(6, 7)]
    values = ["hello,", "world!"]
    line = Line(fields, values)
    fileline = "hello, world!"
    outline = line.write(fileline)
    assert fileline == outline



from typing import IO

import pytest

from cfinterface.components.block import Block
from cfinterface.components.state import ComponentState
from tests.mocks.mock_open import mock_open

from unittest.mock import MagicMock, patch


class DummyBlock(Block):
    BEGIN_PATTERN = "beg"
    END_PATTERN = "end"

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, self.__class__):
            return False
        else:
            return o.data == self.data

    def read(self, file: IO) -> bool:
        self.data = file.readline().strip()
        return True

    def write(self, file: IO) -> bool:
        file.write(self.data)
        return True


def test_single_block_success():
    b1 = Block(state=ComponentState.READ_SUCCESS)
    assert b1.success


def test_single_block_not_found_error():
    b1 = Block(state=ComponentState.NOT_FOUND)
    assert not b1.success


def test_single_block_read_error():
    b1 = Block(state=ComponentState.READ_ERROR)
    assert not b1.success


def test_single_block_properties():
    b1 = Block()
    assert b1.is_first
    assert b1.is_last


def test_block_simple_chain_properties():
    # Build a simple block chain
    b1 = Block()
    b2 = Block()
    b3 = Block()
    # Sets relationships
    b1.next = b2
    b2.previous = b1
    b2.next = b3
    b3.previous = b2
    # Asserts properties
    assert b1.is_first
    assert b3.is_last
    assert not b1.is_last
    assert not b2.is_first
    assert not b2.is_last
    assert not b3.is_first
    assert b1.empty
    assert b2.empty
    assert b3.empty


def test_block_equal_error():
    b1 = Block()
    b2 = Block()
    with pytest.raises(NotImplementedError):
        b1 == b2


def test_block_read_error():
    b = Block()
    with pytest.raises(NotImplementedError):
        m: MagicMock = mock_open(read_data="")
        with patch("builtins.open", m):
            with open("", "r") as fp:
                b.read_block(fp)


def test_block_write_error():
    b = Block(state=ComponentState.READ_SUCCESS)
    with pytest.raises(NotImplementedError):
        m: MagicMock = mock_open(read_data="")
        with patch("builtins.open", m):
            with open("", "r") as fp:
                b.write_block(fp)


def test_dummy_block_equal():
    b1 = DummyBlock()
    b2 = DummyBlock()
    assert b1 == b2


def test_dummy_block_read():
    data = "Hello, world!"
    filedata = (
        "\n".join([DummyBlock.BEGIN_PATTERN, data, DummyBlock.END_PATTERN])
        + "\n"
    )
    m: MagicMock = mock_open(read_data=filedata)
    with patch("builtins.open", m):
        with open("", "r") as fp:
            b = DummyBlock()
            assert DummyBlock.begins(fp.readline())
            b.read_block(fp)
            assert b.data == data
            assert b.success
            assert DummyBlock.ends(fp.readline())


def test_dummy_block_write():
    data = "Hello, world!"
    filedata = ""
    m = mock_open(read_data=filedata)
    with patch("builtins.open", m):
        with open("", "w") as fp:
            b = DummyBlock(state=ComponentState.READ_SUCCESS)
            b.data = data
            b.write_block(fp)
    m().write.assert_called_once_with(data)



from typing import IO, List

from cfinterface.components.block import Block
from cfinterface.reading.blockreading import BlockReading

from tests.mocks.mock_open import mock_open

from unittest.mock import MagicMock, patch


class DummyBlock(Block):
    BEGIN_PATTERN = "beg"
    END_PATTERN = "end"

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, self.__class__):
            return False
        else:
            return o.data == self.data

    def read(self, file: IO) -> bool:
        self.data: List[str] = []
        while True:
            line: str = file.readline()
            self.data.append(line)
            if self.ends(line):
                break
        return True

    def write(self, file: IO) -> bool:
        for line in self.data:
            file.write(line)
        return True


def test_blockreading_empty():
    filedata = ""
    br = BlockReading([DummyBlock])
    m: MagicMock = mock_open(read_data=filedata)
    with patch("builtins.open", m):
        bd = br.read("", "")
        assert br.empty
        assert len(bd) == 1


def test_blockreading_withdata():
    data = "Hello, world!"
    filedata = (
        "\n".join([DummyBlock.BEGIN_PATTERN, data, DummyBlock.END_PATTERN])
        + "\n"
    )
    br = BlockReading([DummyBlock])
    m: MagicMock = mock_open(read_data=filedata)
    with patch("builtins.open", m):
        bd = br.read("", "")
        assert not br.empty
        dbs = [b for b in bd.of_type(DummyBlock)]
        assert len(dbs) == 1
        assert dbs[0].data[0].strip() == DummyBlock.BEGIN_PATTERN
        assert dbs[0].data[1].strip() == data
        assert dbs[0].data[2].strip() == DummyBlock.END_PATTERN



from typing import IO, List, Type
from os.path import join

from cfinterface.components.block import Block
from cfinterface.components.defaultblock import DefaultBlock
from cfinterface.data.blockdata import BlockData


class BlockReading:
    """
    Class for reading custom files based on a BlockData structure.
    """

    def __init__(self, allowed_blocks: List[Type[Block]]) -> None:
        self.__allowed_blocks = allowed_blocks
        self.__data = BlockData(DefaultBlock(data=""))
        self.__last_position_filepointer = 0

    def __read_line_with_backup(self, file: IO) -> str:
        """
        Reads a line of the file, saving the filepointer position
        in case one desired to return to the previous line.

        :param file: THe filepoiner for the reading file
        :type file: IO
        :return: The read line
        :rtype: str
        """
        self.__last_position_filepointer = file.tell()
        return file.readline()

    def __restore_previous_line(self, file: IO):
        """
        Restores the filepointer to the beginning of the previously
        read line.

        :param file: The filepointer
        :type file: IO
        """
        file.seek(self.__last_position_filepointer)

    def __find_starting_block(self, line: str) -> Type[Block]:
        """
        Searches among the given blocks for the block that begins in
        a line of the reading file.

        :param line: A line of the reading file
        :type line: str
        :return: The block type that begins on the given line
        :rtype: Type[Block]
        """
        for b in self.__allowed_blocks:
            if b.begins(line):
                return b
        return DefaultBlock

    def __read_file(self, file: IO) -> BlockData:
        """
        Reads all the blocks from the given blocks in a file and
        returns the BlockData structure.

        :param file: The filepointer
        :type file: IO
        :return: The block data from the file
        :rtype: BlockData
        """
        while True:
            line = self.__read_line_with_backup(file)
            if len(line) == 0:
                break
            self.__restore_previous_line(file)
            blocktype = self.__find_starting_block(line)
            block = blocktype()
            block.read(file)
            self.__data.append(block)
        return self.__data

    def read(self, filename: str, directory: str) -> BlockData:
        """
        Reads a file with a given name in a given directory and
        extracts the data from the specified blocks.

        :param filename: The name of the file
        :type filename: str
        :param directory: The directory where the file is
        :type directory: str
        :return: The data from the blocks found in the file
        :rtype: BlockData
        """
        filepath = join(directory, filename)
        with open(filepath, "r") as fp:
            return self.__read_file(fp)

    @property
    def data(self) -> BlockData:
        return self.__data

    @property
    def empty(self) -> bool:
        return len(self.__data) == 1



from typing import IO
from cfinterface.components.block import Block


class DefaultBlock(Block):
    """
    A class for representing a default block, which contains exactly
    the data from the read line. Mainly used for comments.
    """

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, DefaultBlock):
            return False
        my_data = (self.data, self.previous, self.next)
        o_data = (o.data, o.previous, o.next)
        return my_data == o_data

    def read(self, file: IO) -> bool:
        self.data = file.readline()
        return True

    def write(self, file: IO) -> bool:
        file.write(self.data)
        return True



from abc import abstractmethod
from typing import Any, Optional


class Field:
    """
    Class for representing an field for being read from and written
    in a file.
    """

    def __init__(
        self, size: int, starting_column: int, value: Optional[Any] = None
    ) -> None:
        self._size = size
        self._starting_column = starting_column
        self._ending_column = size + starting_column
        self._value = value

    @abstractmethod
    def read(self, line: str) -> Any:
        """
        Generic method for reading a field from a given line of a file.

        :param line: The line read from the file
        :type line: str
        :return: The value read from the field
        :rtype: Any
        """
        raise NotImplementedError()

    @abstractmethod
    def write(self, line: str) -> str:
        """
        Generic method for writing a field to a given line of a file.

        :param line: The line read from the file
        :type line: str
        :return: The value read from the field
        :rtype: Any
        """
        raise NotImplementedError()

    @property
    def value(self) -> Any:
        return self._value

    @value.setter
    def value(self, val: Any):
        self._value = val



from typing import List, Type

from cfinterface.components.block import Block
from cfinterface.components.defaultblock import DefaultBlock
from cfinterface.data.blockdata import BlockData
from cfinterface.reading.blockreading import BlockReading
from cfinterface.writing.blockwriting import BlockWriting


class BlockFile:
    """
    Class that models a file divided by blocks, where the reading
    and writing are given by a series of blocks.
    """

    BLOCKS: List[Type[Block]] = []

    def __init__(
        self,
        data=BlockData(DefaultBlock("")),
    ) -> None:
        self.__data = data

    @classmethod
    def read(cls, directory: str, filename: str = ""):
        """
        Reads the blockfile data from a given file in disk.

        :param filename: The file name in disk
        :type filename: str
        :param directory: The directory where the file is
        :type directory: str
        """
        reader = BlockReading(cls.BLOCKS)
        return cls(reader.read(filename, directory))

    def write(self, directory: str, filename: str = ""):
        """
        Write the blockfile data to a given file in disk.

        :param filename: The file name in disk
        :type filename: str
        :param directory: The directory where the file will be
        :type directory: str
        """
        writer = BlockWriting(self.__data)
        writer.write(filename, directory)

    @property
    def data(self) -> BlockData:
        return self.__data



from cfinterface.components.block import Block
from cfinterface.data.blockdata import BlockData


class DummyBlock(Block):
    BEGIN_PATTERN = "beg"
    END_PATTERN = "end"

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, self.__class__):
            return False
        else:
            return o.data == self.data


def test_blockdata_append():
    bd = BlockData(DummyBlock(data=-1))
    n_blocks = 10
    for i in range(n_blocks):
        bd.append(DummyBlock(data=i))
    assert len(bd) == n_blocks + 1
    assert bd.first.data == -1
    assert bd.last.data == n_blocks - 1


def test_blockdata_preppend():
    bd = BlockData(DummyBlock(data=-1))
    n_blocks = 10
    for i in range(n_blocks):
        bd.preppend(DummyBlock(data=i))
    assert len(bd) == n_blocks + 1
    assert bd.first.data == n_blocks - 1
    assert bd.last.data == -1


def test_blockdata_add_before():
    bd = BlockData(DummyBlock(data=-1))
    b1 = DummyBlock(data=1)
    bd.append(b1)
    b2 = DummyBlock(data=2)
    bd.add_before(b1, b2)
    assert b1.previous == b2
    assert b2.next == b1


def test_blockdata_add_after():
    bd = BlockData(DummyBlock(data=-1))
    b1 = DummyBlock(data=1)
    bd.append(b1)
    b2 = DummyBlock(data=2)
    bd.add_after(b1, b2)
    assert b1.next == b2
    assert b2.previous == b1


def test_blockdata_remove():
    bd = BlockData(DummyBlock(data=-1))
    b1 = DummyBlock(data=1)
    bd.append(b1)
    assert len(bd) == 2
    bd.remove(b1)
    assert len(bd) == 1


def test_blockdata_of_type():
    bd = BlockData(DummyBlock())
    bd.append(Block())
    assert len(bd) == 2
    assert len([b for b in bd.of_type(DummyBlock)]) == 1
